library jsonify;

import 'dart:mirrors';

T createFromJson<T>(Map<String, dynamic> json) {
  // FIXME: we must make sure that all final fields are initialized
  final ClassMirror classMirror = reflectType(T);
  if (json == null) classMirror.newInstance(Symbol.empty, [], {});
  final Map<Symbol, DeclarationMirror> declarations = classMirror.declarations;
  final Map<Symbol, dynamic> data = Map<Symbol, dynamic>();
  // Convert strings to symbols
  for (final MapEntry<String, dynamic> entry in json.entries) {
    Symbol symbol = Symbol(entry.key);
    if (declarations.containsKey(symbol)) {
      final VariableMirror declaration = declarations[symbol];
      if (declaration.isPrivate) continue;
      dynamic value = entry.value;
      InstanceMirror valueMirror = reflect(entry.value);
      TypeMirror valueType = valueMirror.type;
      if (valueType.isAssignableTo(declaration.type)) {
        // If the value is directly assignable then let's do it
        data[symbol] = value;
      } else if (value.runtimeType == json.runtimeType) {
        // This means that it's another map, so let's try
        // to run this very function on the and see how it
        // goes
        ClassMirror targetClassMirror = declaration.type;
        data[symbol] = targetClassMirror.newInstance(
          Symbol("fromJson"),
          [value],
        ).reflectee;
      } else if (value != "values") {
        Symbol valueAsSymbol = Symbol(value);
        ClassMirror enumMirror = declaration.type;
        if (!enumMirror.isEnum) {
          throw ('cannot initialize ' +
              '`${entry.key}\' because I don\'t know ' +
              'the type, it has a value though: $value');
        }
        Map<Symbol, MethodMirror> staticMembers = enumMirror.staticMembers;
        if (staticMembers.containsKey(valueAsSymbol)) {
          MethodMirror methodMirror = staticMembers[valueAsSymbol];
          if (methodMirror.isGetter && methodMirror.isSynthetic) {
            data[symbol] = enumMirror.getField(valueAsSymbol).reflectee;
          }
        }
      } else {
        throw ('cannot initialize ' +
            '`${entry.key}\' because I don\'t know ' +
            'the type, it has a value though: $value');
      }
    } else {
      print(
        'ignoring field `${entry.key}\' because it\'s not a field of `$T\'',
      );
    }
  }
  return classMirror.newInstance(Symbol.empty, [], data).reflectee;
}
